/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

import com.ibm.rqm.adapter.library.connector.AbstractTestToolAdapter;
import com.ibm.rqm.adapter.library.data.AbstractLogEvent;
import com.ibm.rqm.adapter.library.data.CustomProperty;
import com.ibm.rqm.adapter.library.data.ElementTypeEnum;
import com.ibm.rqm.adapter.library.data.ILogEvent;
import com.ibm.rqm.adapter.library.data.ILogResult;
import com.ibm.rqm.adapter.library.data.ResultStateEnum;


/**
 * The entity which will execute the tasks which has been taken up by the
 * adapter.<br> 
 * The tasks that the adapter can perform are registered with the
 * adapter web plug-in registry as capabilities <br>( extension point
 * <code>com.ibm.rqm.planning.common.executionScriptType</code><br>
 * property <code>capability</code>)
 * <br>
 * eg for plug-in registry.<br> 
 * <code>
 * &lt;extension
 *     point="com.ibm.rqm.planning.common.executionScriptType"&gt;
 *  		&lt;type
 *          adapterName="Sample: Command Line Adapter"
 *          description="RQM-KEY-CMD-APTR-TYPE-DESC"
 *          id="com.ibm.rqm.adapter.commandline"
 *          name="RQM-KEY-CMD-APTR-TYPE-NAME"&gt;
 *       &lt;capability
 *             name="EXECUTE"&gt;
 *       &lt;/capability&gt;
 *       &lt;capability
 *             name="IMPORT"&gt;
 *       &lt;/capability&gt;
 *   		&lt;/type&gt;
 *	&lt;/extension&gt;
 * </code>
 * <br>
 * For a capability <code>IMPORT</code> the method to implement is {@link IQMToolAdapter#getResources(String)}
 * For a capability <code>EXECUTE</code> the method to implement is {@link IQMToolAdapter#runTest(String[])}
 * 
 * Note: At present (in 1.0) these are the only two capabilities supported by the framework
 * 
 * @author amathur
 */
public class CommandLineAdapter extends AbstractTestToolAdapter {

	private File stdOutFile = null;
	private File stdErrFile = null;
	private String errorMessage = "";
	private Process process = null;
	private TrackProgress trackProgress= null;
	
	private LogProgress logProgress = null;
	
	public static final int STATUS_NOTSTARTED = 0;
	public static final int STATUS_INITIALIZED = 1;
	public static final int STATUS_EXECUTING = 2;
	public static final int STATUS_EXECUTION_DONE = 3;
	public static final int STATUS_DONE = 4;
	public static final int STATUS_STOPPED = 5;
	
	int status = STATUS_NOTSTARTED;
	
	/**
	 * Import is not supported in the Command line adapter. All extensions
	 * should override this method to provide support for import.
	 * 
	 * @param path -
	 *            The input location on the adapter machine
	 * @return - an array of strings with the name of resources.
	 */
	public String[] getResources(String path) {
		return null;
	}
	
	public static class MyFd extends Thread {
		private InputStream fd;
		private File f;
		public MyFd(InputStream fd,File f) {
			this.fd=fd;
			this.f=f;
		}
		public void run() {
			byte[] bytes = new byte[4096];
			try {
				BufferedInputStream bis = new BufferedInputStream(fd);
				BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(f));
				int amount=0;
				while ((amount=bis.read(bytes)) >= 0) {
		        	out.write(bytes,0,amount);
		        }
				bis.close();
				out.close();
				fd.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	public void setErrorMessage(String mesg)
	{
		errorMessage = mesg;
	}
	
	public String getErrorMessage()
	{
		return errorMessage;
	}
	
	/* (non-Javadoc)
	 * @see com.ibm.rqm.adapter.library.connector.IQMToolAdapter#runTest(java.lang.String[])
	 */
	public void runTest(String[] args) {
		runTest(args, null);
	}
	
	public void runTest(String[] args, Properties executionVariables) {
		setStatus(STATUS_NOTSTARTED);
		StringBuffer commandLine = new StringBuffer();
		if ('\\' == File.separatorChar){
			commandLine.append("\"");
			commandLine.append(args[0]);
			commandLine.append("\"");
		} else {
			commandLine.append(args[0]);
		}
		if (args[1] != null) {
			commandLine.append(" ");
			commandLine.append(args[1]);
		}
		
		try {
			List<AbstractLogEvent> logEvents = new ArrayList<AbstractLogEvent>();
			
			trackProgress = new TrackProgress();
			int exitValue = 0;
			AbstractLogEvent event = new CLALogEvent();
			setStatus(STATUS_INITIALIZED);
			try {
				trackProgress.start();
				createOutFiles();
				event.starttime = System.currentTimeMillis();
				// Start command line execution and track progress ...
				process = startProcess(commandLine.toString(), executionVariables);
				trackProgress.setProcess(process);
				MyFd outFd=new MyFd(process.getInputStream(),stdOutFile);outFd.start();
				MyFd errFd=new MyFd(process.getErrorStream(),stdErrFile);errFd.start();
				setStatus(STATUS_EXECUTING);
				exitValue = process.waitFor();
				setStatus(STATUS_EXECUTION_DONE);
				outFd.join();
				errFd.join();
				trackProgress.stop();
				System.out.println("ExitCode: "+exitValue); // show result code in console!!!!
				ResultStateEnum result = CommandLinePreferences.prefs().getExitCode(exitValue);
				((CommandLineLogResult)logResult).setStatus(result!= null ? result : ResultStateEnum.UNKNOWN);
				// Use notes to add link to any external data or testers note about execution
//				((CommandLineLogResult)logResult).setNotes("<a href=\"http://google.com\">google</a><br/><br/><br/>");
				// Load additional system properties as needed
				((CommandLineLogResult)logResult).setCustomProperties(getCustomProperties());
				((CommandLineLogResult)logResult).setExecutionVariables(executionVariables);
			} catch (Exception e) {
				if (this.status == STATUS_STOPPED)
					throw e;
				exitValue = 2;
				Logger.Log.error(e.getMessage(), e);		
				((CommandLineLogResult)logResult).setStatus(ResultStateEnum.ERROR);
				((CommandLineLogResult)logResult).setNotes(e.getMessage());
				MyFd errFd=new MyFd(new ByteArrayInputStream((e.getMessage()+getErrorMessage()).getBytes("UTF-8")),stdErrFile);errFd.start();
				errFd.join();
			} finally {
				trackProgress.stop();
			}
			
			event.resultState = exitValue == 0 ? ResultStateEnum.PASSED : (exitValue == 1 ? ResultStateEnum.FAILED : ResultStateEnum.ERROR);
			event.type = ElementTypeEnum.REPORTING;
			event.endtime = System.currentTimeMillis();
			logEvents.add(event);

			((CommandLineLogResult)logResult).setLogEvents(logEvents.toArray(new ILogEvent[] {}));
		} catch (Exception e) {
			Logger.Log.error(e.getMessage(), e);		
			((CommandLineLogResult)logResult).setStatus(ResultStateEnum.ERROR);
		}
	}

	public void createOutFiles() throws IOException {
		stdOutFile=File.createTempFile("QMAdapter",".out");  //$NON-NLS-1$//$NON-NLS-2$
		stdErrFile=File.createTempFile("QMAdapter",".err");  //$NON-NLS-1$//$NON-NLS-2$
	}
	
	public void cleanupOutFiles() {
		if(stdOutFile!=null) {
			stdOutFile.delete();
		}
		if(stdErrFile!=null) {
			stdErrFile.delete();
		}
	}
	
	public void setLogResult(ILogResult logResult){
		this.logResult = logResult;
	}

	public void setLogProgress(LogProgress logProgress){
		this.logProgress = logProgress;
	}

	public File getStdOutFile() {
		return stdOutFile;
	}

	public File getStdErrFile() {
		return stdErrFile;
	}
	
	public boolean uploadError(String errorMessage, ILogResult logResult){
		ByteArrayInputStream bais = null;
		byte[] messageBytes;
		try{
			try {
				messageBytes = errorMessage.toString().getBytes("UTF-8");
			} catch (UnsupportedEncodingException e1) {
				messageBytes = errorMessage.toString().getBytes();
			}
			bais = new ByteArrayInputStream(messageBytes);
			File stdErrFile = getStdErrFile();
			if (stdErrFile == null){
				createOutFiles();
				stdErrFile = getStdErrFile();
			}
			CommandLineAdapter.MyFd stdErrorThread = new CommandLineAdapter.MyFd(bais, stdErrFile);stdErrorThread.start();
			stdErrorThread.join();
			try{
				if (stdErrFile != null && stdErrFile.getCanonicalFile().length() != 0){
					logResult.addAttachment(stdErrFile.getCanonicalPath(), "text/plain"); //$NON-NLS-1$
				}
			}
			catch(IOException e){
				Logger.Log.error(e.getMessage(),e);
			}
			return true;
		} catch (Exception e3){
			return false;
		} 
	}
	
	/*
	 * Get the environment variables to be used in the presence of execution variables.
	 */
	private Process startProcess(String commandline, Properties vars) throws IOException {
		StringTokenizer st = new StringTokenizer(commandline);
		String[] cmd = new String[st.countTokens()];
	 	for (int i = 0; st.hasMoreTokens(); i++)
		    cmd[i] = st.nextToken();
		ProcessBuilder pb = new ProcessBuilder(cmd);
		if ( vars != null ) {
			Map<String, String> env = pb.environment();
			Enumeration<Object> keys = vars.keys();
			while (keys.hasMoreElements()) {
				String key = (String)keys.nextElement();
				String value = (String)vars.getProperty(key);
				env.put("qm."+cleanPropertyName(key), value); //$NON-NLS-1$
			}
		}
		return pb.start();
	}
	
	private String cleanPropertyName(String name) {
		char[] chars = name.toCharArray();
		int length = (chars != null ? chars.length : 0);
		for (int i = 0; i < length; ++i) {
			if (!Character.isJavaIdentifierPart(chars[i]))
				chars[i] = '_';
		}
		return new String(chars);
	}
	
	/*
	 * Get custom properties provided by the user in the preferences file and optionally add in
	 * any system properties that are requested.
	 */
	private CustomProperty[] getCustomProperties() {
		Map<String,String> props = CommandLinePreferences.prefs().getNonRqmProperties();
		if (CommandLinePreferences.prefs().getBooleanProperty(CommandLinePreferences.INCLUDE_SYSTEM_PROPERTIES, false)) {
			Properties systemProps = System.getProperties();
			Enumeration<Object> keys = systemProps.keys();
			while (keys.hasMoreElements()) {
				String key = (String)keys.nextElement();
				if ( key.startsWith("os.") || key.startsWith("user.") )
					props.put(key, systemProps.getProperty(key));
			}
		}
		CustomProperty[] result = new CustomProperty[props.size()];
		Set<String> keys = props.keySet();
		Iterator<String> iter = keys.iterator();
		int index = 0;
		while (iter.hasNext()) {
			String key = iter.next();
			CustomProperty prop = new CustomProperty();
			prop.setPropertyName(key);
			String value = props.get(key);
			value.replaceAll("<", "&lt;");
			value.replaceAll(">", "&gt;");
			prop.setPropertyValue(value);
			//System.out.println(key +" : "+value);
			prop.setNameSpaceWithPrefix("rqmprop", "http://jazz.net/xmlns/alm/rqm/common");
			result[index++] = prop;
		}
		return result;
	}
	
	public void stopExecution(){
		
		
		setStatus(STATUS_STOPPED);
		
		if(process!= null)
			process.destroy();
		
    	if(this.trackProgress!=null)
 		   trackProgress.stop();	
    	
    	
		
	}
	
	public synchronized int setStatus(int setStatus)
	{
		if (this.status == STATUS_EXECUTION_DONE)
			return this.status;
		
		if (setStatus == STATUS_STOPPED)
		{
			this.status = setStatus;
			return status;
		}
		
		if (this.status == STATUS_STOPPED)
			throw new ExecutionStoppedException();
		
		this.status = setStatus;
		return setStatus;
			
	}
	
	private class TrackProgress {
		
		Thread thread = null;
		boolean running = false;
		Process process = null;
		
		public synchronized Thread start() {
			if (!running) {
				running = true;
				thread = new Thread(new Runnable() {
					public void run() {
						monitorExecution();
						running = false;
					}
				});
				thread.setDaemon(true);
				thread.setName("CommandLineAdapter.TrackProgress");
				thread.start();
			}
			return thread;
		}
		
		public synchronized void stop() {
			if ( running ) {
				running = false;
				thread.interrupt();
			}
		}
		
		public void setProcess(Process process) {
			this.process = process;
		}
		
		private void monitorExecution() {
			String monitorProperties = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.EXECUTION_PROPERTIES, "");
			if ( logProgress != null && monitorProperties != null && !monitorProperties.equals("") ) {
				System.out.println("Monitor Progress: ["+monitorProperties+"]");
				File monitor = new File(monitorProperties);
				long since = -1;
				while (running) {
					try {
						// start monitoring after first polling interval completes ....
						Thread.sleep(5000); // use normal ping interval?					

						//process.
						if ( process == null )
							continue; // process has not been started yet ...
						
						// Check if there is a status file being written by the application
						// under test and if so if it has changed.  If changed then load and
						// update the current execution status.
						if ( monitor.exists() && since != monitor.lastModified() ) {
							Properties props = new Properties();
							try {
								FileInputStream stream = new FileInputStream(monitor);
								props.load(stream);
								stream.close();
								String comment = "<progress must be in 0..100>";
								int progress = parseInt(props.getProperty("progress", "0"));
								if ( progress >= 0 && progress <= 100 ) {
									comment = props.getProperty("comment", "");
									logProgress.logProgress(progress, comment, null);
								}
								System.out.println("Progress: "+progress+": "+comment);
							} catch (IOException ioexc) {
								System.out.println(ioexc.toString());
							}
							since = monitor.lastModified();
						}
					} catch (InterruptedException e) {
						System.out.println("Monitor thread interrupted");
					}
				}
			}
		}
		
		private boolean isProcessAlive(Process process) {
			if ( process == null ) return true;  // too soon, not registered yet ...
			boolean result = true;
			try {
				process.exitValue(); // throws IllegalThreadStateException if halted ...
			} catch ( IllegalThreadStateException itse ) {
				result = false;
			}
			return result;
		}
		
		private int parseInt(String number) {
			int result = 0;
			try {
				result = (number != null ? Integer.parseInt(number.trim()) : 0);
			} catch (NumberFormatException numFormat) {
			}
			return result;
		}
		
	}
	
}
